#!/usr/bin/env python

"""
Moin wiki format parser.

Copyright (C) 2012, 2013, 2015, 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from cgi import escape
import re

# Regular expressions.

syntax = {
    # Page regions:
    "marker"        : (r"^\s*([{]{3,}|[}]{3,})",    re.MULTILINE | re.DOTALL),  # {{{... or }}}...

    # Region contents:
    "header"        : (r"#!(.*?)\n",                0),                         # #! char-excl-nl
    "break"         : (r"^\s*?\n",                  re.MULTILINE),              # blank line
    }

# Define patterns for the regular expressions.

patterns = {}
for name, (value, flags) in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | flags)



# Document nodes.

class Container:

    "A container of document nodes."

    def __init__(self, nodes):
        self.nodes = nodes

    def append(self, node):
        self.nodes.append(node)

    def normalise(self):

        "Combine adjacent text nodes."

        nodes = self.nodes
        self.nodes = []
        text = None

        for node in nodes:

            # Open a text node or merge text into an open node.

            if isinstance(node, Text):
                if not text:
                    text = node
                else:
                    text.merge(node)

            # Close any open text node and append the current node.

            else:
                if text:
                    self.append(text)
                    text = None
                self.append(node)

        # Add any open text node.

        if text:
            self.append(text)

    def __str__(self):
        return self.prettyprint()

    def prettyprint(self, indent=""):
        pass

class Region(Container):

    "A region of the page."

    transparent_region_types = ["wiki"]

    def __init__(self, nodes, level=0, type=None):
        Container.__init__(self, nodes)
        self.level = level
        self.type = type

    def have_start(self, s):
        return self.is_transparent() and s.startswith("{")

    def have_end(self, s):
        return self.level and s.startswith("}") and self.level == len(s)

    def is_transparent(self):
        return not self.level or self.type in self.transparent_region_types

    def __repr__(self):
        return "Region(%r, %r, %r)" % (self.nodes, self.level, self.type)

    def prettyprint(self, indent=""):
        l = ["%sRegion: level=%d type=%s" % (indent, self.level, self.type)]
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def to_string(self, out):
        out.start_region(self.level, self.type)
        for node in self.nodes:
            node.to_string(out)
        out.end_region(self.level, self.type)

class Block(Container):

    "A block in the page."

    def __init__(self, nodes, final=True):
        Container.__init__(self, nodes)
        self.final = final

    def __repr__(self):
        return "Block(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sBlock: final=%s" % (indent, self.final)]
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def to_string(self, out):
        out.start_block(self.final)
        for node in self.nodes:
            node.to_string(out)
        out.end_block(self.final)

class Text:

    "A text node."

    def __init__(self, s):
        self.s = s

    def merge(self, text):
        self.s += text.s

    def __repr__(self):
        return "Text(%r)" % self.s

    def prettyprint(self, indent=""):
        return "%sText: %r" % (indent, self.s)

    def to_string(self, out):
        out.text(self.s)



# Serialisation.

class Serialiser:

    "General serialisation support."

    def __init__(self, out):
        self.out = out

class MoinSerialiser(Serialiser):

    "Serialisation of the page."

    def start_region(self, level, type):
        out = self.out
        if level:
            out("{" * level)        # marker
        if type and level:
            out("#!%s\n" % type)    # header

    def end_region(self, level, type):
        out = self.out
        if level:
            out("}" * level)        # marker

    def start_block(self, final):
        pass

    def end_block(self, final):
        if not final:
            self.out("\n")

    def text(self, s):
        self.out(s)

class HTMLSerialiser(Serialiser):

    "Serialisation of the page."

    def start_region(self, level, type):
        l = []
        out = l.append
        if level:
            out("level-%d" % level)                 # marker

        # NOTE: Encode type details for CSS.

        if type:
            out("type-%s" % escape(type, True))     # header

        self.out("<span class='%s'>" % " ".join(l))

    def end_region(self, level, type):
        self.out("</span>")

    def start_block(self, final):
        self.out("<p>")

    def end_block(self, final):
        self.out("</p>")

    def text(self, s):
        self.out(escape(s))



# Tokenising functions.

class TokenStream:

    "A stream of tokens taken from a string."

    def __init__(self, s):
        self.s = s
        self.pos = 0
        self.match = None
        self.matching = None

    def read_until(self, pattern_names, remaining=True):

        """
        Find the first match for the given 'pattern_names'. Return the text
        preceding any match, the remaining text if no match was found, or None
        if no match was found and 'remaining' is given as a false value.
        """

        first = None
        self.matching = None

        # Find the first matching pattern.

        for pattern_name in pattern_names:
            match = patterns[pattern_name].search(self.s, self.pos)
            if match:
                start, end = match.span()
                if self.matching is None or start < first:
                    first = start
                    self.matching = pattern_name
                    self.match = match

        if self.matching is None:
            if remaining:
                return self.s[self.pos:]
            else:
                return None
        else:
            return self.s[self.pos:first]

    def read_match(self):

        "Return the matched text, updating the position in the stream."

        if self.match:
            _start, self.pos = self.match.span()
            s = self.match.group(1)
            self.match = None
            return s
        else:
            self.pos = len(self.s)
            return None



# Parser functions.

def parse_page(s):

    """
    Parse page text 's'. Pages consist of regions delimited by markers.
    """

    items = TokenStream(s)

    # Define a region for the page and parse it.

    region = Region([])
    parse_region(items, region)
    return region

def parse_region(items, region):

    "Parse the data provided by 'items' to populate 'region'."

    # Parse section headers.

    parse_region_header(items, region)

    if region.is_transparent():
        parse_region_wiki(items, region)
    else:
        parse_region_opaque(items, region)

def parse_region_wiki(items, region):

    "Parse the data provided by 'items' to populate a wiki 'region'."

    # Process exposed text and sections.

    block = Block([])
    region.append(block)

    while True:

        # Obtain text before any marker or the end of the input.

        match_text = items.read_until(["break", "marker"])
        if match_text:
            block.append(Text(match_text))

        # Obtain any feature.

        feature = items.read_match()

        # End of input.

        if not items.matching:
            break

        # Start a section if an appropriate marker is given.

        if region.have_start(feature):

            # Define the section and parse it.

            _region = Region([], len(feature))
            region.append(_region)
            parse_region(items, _region)

            # Start a new block after the section.

            block = Block([])
            region.append(block)

        # Interpret the given marker, closing the current section if the
        # given marker is the corresponding end marker for the current
        # section.

        elif region.have_end(feature):
            break

        # Start a new block if a paragraph break is found.

        elif items.matching == "break":
            block.final = False
            block = Block([])
            region.append(block)

        # Add any inappropriate marker to the text.

        else:
            block.append(Text(feature))

    region.normalise()

def parse_region_opaque(items, region):

    "Parse the data provided by 'items' to populate an opaque 'region'."

    # Process exposed text and sections.

    while True:

        # Obtain text before any marker or the end of the input.

        match_text = items.read_until(["marker"])
        if match_text:
            region.append(Text(match_text))

        # Obtain any marker.

        marker = items.read_match()

        # End of input.

        if not marker:
            break

        # Interpret the given marker, closing the current section if the
        # given marker is the corresponding end marker for the current
        # section.

        if region.have_end(marker):
            break

        # Add any inappropriate marker to the text.

        else:
            region.append(Text(marker))

    region.normalise()

def parse_region_header(items, region):

    """
    Parse the region header from the 'items', setting it for the given 'region'.
    """

    if items.read_until(["header"], False) == "": # None means no header
        region.type = items.read_match()



# Top-level functions.

parse = parse_page

def serialise(doc, serialiser=MoinSerialiser):
    l = []
    doc.to_string(serialiser(l.append))
    return "".join(l)

# vim: tabstop=4 expandtab shiftwidth=4
